import numpy as np
import pytest
from classrad.utils.utils import get_peak_from_histogram

np.random.seed(9)


def test_get_peak_from_histogram():
    uniform = np.array([i for i in range(1000)])
    gaussian = np.random.normal(0, 100, 1000)
    steps = np.arange(-100, 100, step=10)
    uniform_bins, uniform_bin_edges = np.histogram(uniform, bins=steps)
    assert get_peak_from_histogram(uniform_bins, uniform_bin_edges) == 5
    gaussian_bins, gaussian_bin_edges = np.histogram(gaussian, bins=steps)
    assert abs(get_peak_from_histogram(gaussian_bins, gaussian_bin_edges)) < 10
    with pytest.raises(AssertionError):
        get_peak_from_histogram(bins=[], bin_edges=[])



import collections
from pathlib import Path

import streamlit as st
import utils


def set_up_templates():
    template_dict = collections.defaultdict(dict)
    templates = Path("./templates").rglob("*.py")
    templates = sorted(templates, key=lambda e: e.name)
    for template in templates:
        try:
            # Templates with task + framework.
            workflow = template.parent.name
            task = template.stem
            template_dict[workflow][task] = template
        except ValueError:
            template_dict[template.name] = template
    return template_dict


def main():
    template_dict = set_up_templates()
    st.set_page_config(layout="wide")
    col1, col2 = st.columns(2)
    with col1:
        st.title("Classy Radiomics")
    with col2:
        st.write(
            """
        ####
        The easiest framework for training models
        using `pyradiomics` and `scikit-learn`.
        """
        )

    with st.sidebar:
        st.write("## Task")
        task = st.selectbox("Select workflow", list(template_dict.keys()))
        if isinstance(template_dict[task], dict):
            framework = st.selectbox(
                "Select task", list(template_dict[task].keys())
            )
            template_path = template_dict[task][framework]
        else:
            template_path = template_dict[task]

    # Show template-specific sidebar components
    # (based on sidebar.py in the template dir).
    template = utils.import_from_file("template", template_path)
    template.show()


if __name__ == "__main__":
    main()



from multiprocessing import Pool

import pandas as pd
from classrad.feature_extraction.extractor import FeatureExtractor
from classrad.utils.utils import time_it


class StreamlitFeatureExtractor(FeatureExtractor):
    @time_it
    def extract_features(self, feature_set=None, verbose=None, progressbar=None):
        """
        Extract features from a set of images.
        """
        if feature_set is not None:
            self.feature_set = feature_set
        if verbose is not None:
            self.verbose = verbose

        # Get the feature extractor
        self.logger.info("Initializing feature extractor")
        self.get_feature_extractor()

        self.logger.info("Initializing feature dataframe")
        self.initialize_feature_df()

        # Get the feature values
        self.logger.info("Extracting features")
        if self.num_threads is not None:
            self.get_features_parallel(progressbar=progressbar)
        else:
            self.get_features(progressbar=progressbar)
        self.save_feature_df()
        return self.feature_df

    def get_features(self, progressbar=None):
        """
        Get the feature values
        """
        rows = list(self.df.iterrows())
        for i, (index, row) in enumerate(rows):
            feature_series = self.add_features_for_single_case(row)
            self.feature_df = self.feature_df.append(feature_series, ignore_index=True)
            if progressbar:
                fraction_complete = (i + 1) / len(rows)
                progressbar.progress(fraction_complete)
        return self

    def get_features_parallel(self, progressbar=None):
        try:
            _, df_rows = zip(*self.df.iterrows())
            n = len(self.df)
            p = Pool(self.num_threads)
            if progressbar:
                results = []
                for i, result in enumerate(
                    p.imap(self.add_features_for_single_case, df_rows)
                ):
                    results.append(result)
                    fraction_complete = (i + 1) / n
                    progressbar.progress(fraction_complete)
            else:
                results = p.map(self.add_features_for_single_case, df_rows)
            self.feature_df = pd.concat(results, axis=1).T
        except Exception:
            print("Multiprocessing failed! :/")
        return self



from pathlib import Path
import pandas as pd
import streamlit as st


def show():
    st.write(
        """
    #####
    Before you proceed with the next tasks you need to create a .csv table.
    The table should contain the following information for each case:
    - **ID** (has to be unique)
    - Path to the **image**
    - Path to the **mask**
    - **Label** (0 or 1)

    Example:
    """
    )
    example_path_df = pd.read_csv("paths_example.csv")
    st.write(example_path_df)
    st.write(
        """
    #####
    You can create the table manually (e.g. in Excel) or, if your dataset
    follows the specified structure, you can generate it below.
    """
    )
    dir_structures = {
        "Separate folder for each case": """
            Expected format:
            ```
            <dataset folder>
            +-- case1
            |   +-- image.nii.gz
            |   +-- segmentation.nii.gz
            +-- case2
            |   +-- image.nii.gz
            |   +-- segmentation.nii.gz
            ...
            ```
            """,
        "One folder for all the images, and one for the segmentations": """
            Expected format:
            ```
            <dataset folder>
            +-- images
            |   +-- 1.nii.gz
            |   +-- 2.nii.gz
            |   +-- ...
            +-- segmentations
            |   +-- 1.nii.gz
            |   +-- 2.nii.gz
            |   +-- ...
            ...
            ```
            """,
    }
    col1, col2 = st.columns(2)
    with col1:
        format = st.selectbox("Dataset structure", dir_structures.keys())
    with col2:
        st.write(dir_structures[format])
    data_dir = st.text_input(
        "Enter absolute path to the folder with your dataset:"
    )
    if not data_dir:
        st.stop()
    data_dir = Path(data_dir).absolute()
    if not data_dir.is_dir():
        st.error(f"The entered path is not a directory ({data_dir})")
    else:
        st.success(f"The entered path is a directory. ({data_dir})")


if __name__ == "__main__":
    show()



import json
import os
import shutil


def make_if_dont_exist(folder_path, overwrite=False):
    """
    creates a folder if it does not exists
    input:
    folder_path : relative path of the folder which needs to be created
    over_write :(default: False) if True overwrite the existing folder
    """
    if os.path.exists(folder_path):

        if not overwrite:
            print(f"{folder_path} exists.")
        else:
            print(f"{folder_path} overwritten.")
            shutil.rmtree(folder_path)
            os.makedirs(folder_path)

    else:
        os.makedirs(folder_path)
        print(f"{folder_path} created!")


def load_json(file_name):
    with open(file_name) as f:
        data = json.load(f)
    return data


def save_json(data, file_name):
    with open(file_name, "w") as f:
        json.dump(data, f, indent=4)



import os
from pathlib import Path
import streamlit as st
import utils
import json
from classrad.config import config


def radiomics_params():
    config_dir = os.path.dirname(config.__file__)
    param_dir = Path(config_dir) / "pyradiomics_params"
    presets = {"CT from Baessler et al. (2019)": "Baessler_CT.yaml"}
    preset_options = list(presets.keys())
    name = st.selectbox("Choose a preset", preset_options)
    setup = utils.read_yaml(param_dir / presets[name])
    st.write(""" Filters: """)
    filter = setup["imageType"]
    col1, col2, col3 = st.columns(3)
    with col1:
        turn_on_original = st.checkbox("original", value=("Original" in filter))
    with col2:
        turn_on_log = st.checkbox("Laplacian of Gaussian", value=("LoG" in filter))
        if turn_on_log:
            sigmas = filter["LoG"]["sigma"]
            for sigma in sigmas:
                st.number_input("sigma", value=sigma)
    with col3:
        turn_on_wavelet = st.checkbox("Wavelet", value=("Wavelet" in filter))
    classes = setup["featureClass"]
    all_classes = ["firstorder", "shape", "glcm", "glszm", "glrlm", "gldm"]
    st.write(""" Classes: """)
    cols = st.columns(len(all_classes))
    for i in range(len(all_classes)):
        with cols[i]:
            st.checkbox(all_classes[i], value=True)
    with cols[5]:
        st.checkbox("glcm", value=True)
    setting = setup["setting"]
    st.write(""" Normalization: """)
    normalization = setting["normalize"]
    st.checkbox("Normalize", value=normalization)
    st.number_input("Bin width", value=setting["binWidth"])
    st.write(""" Full parameter file: """, setup)


def choose_preset():
    config_dir = os.path.dirname(config.__file__)
    param_dir = Path(config_dir) / "pyradiomics_params"
    presets = config.PRESETS
    preset_options = list(presets.keys())
    name = st.selectbox(
        "Choose a preset with parameters for feature extraction", preset_options
    )
    preset_setup = utils.read_yaml(param_dir / presets[name])
    return preset_setup


def update_setup_with_class(setup, class_name, class_active):
    assert "featureClass" in setup
    if class_active[class_name]:
        setup["featureClass"][class_name] = []
    else:
        setup["featureClass"].pop(class_name, None)
    return setup


def select_classes(preset_setup, final_setup):
    st.write(""" #### Select classes: """)

    all_feature_names = config.PYRADIOMICS_FEATURE_NAMES
    all_classes = all_feature_names.keys()
    preset_classes = preset_setup["featureClass"]
    class_active = {}
    cols = st.columns(len(all_classes))
    for i, class_name in enumerate(all_classes):
        with cols[i]:
            class_active[class_name] = st.checkbox(
                class_name, value=(class_name in preset_classes), key=i
            )
        final_setup = update_setup_with_class(final_setup, class_name, class_active)
    # st.write("##### Optional: select specific features within class:")
    st.info(
        """
        If you select a class, all its features are included by default.
        You can however select only specific features (optional):
        """
    )
    cols = st.columns(len(preset_classes.keys()))
    for i, class_name in enumerate(preset_classes.keys()):
        with cols[i]:
            if class_active[class_name]:
                feature_names = st.multiselect(
                    class_name, all_feature_names[class_name], key=class_name
                )
                final_setup["featureClass"][class_name] = feature_names
    return final_setup


def radiomics_params_voxelbased():
    param_dir = Path(config.PARAM_DIR)
    presets = config.PRESETS
    preset_options = list(presets.keys())
    name = st.selectbox(
        "Choose a preset with parameters for feature extraction", preset_options
    )
    preset_setup = utils.read_yaml(param_dir / presets[name])
    final_setup = preset_setup.copy()

    with st.expander("Manually edit the extraction parameters"):
        final_setup = select_classes(preset_setup, final_setup)

        setting = preset_setup["setting"]
        col1, col2, col3 = st.columns(3)
        with col1:
            st.write(""" ##### Normalization: """)
            normalization = setting["normalize"]
            is_normalized = st.checkbox("Normalize", value=normalization)
            if is_normalized:
                final_setup["setting"]["normalize"] = True
            else:
                final_setup["setting"]["normalize"] = False
        with col2:
            bin_width = st.number_input("""Bin width:""", value=setting["binWidth"])
            final_setup["setting"]["binWidth"] = bin_width
        with col3:
            label = st.number_input("Label:", value=setting["label"])
            final_setup["setting"]["label"] = int(label)
        st.write(""" #### Full parameter file: """, preset_setup)

    return final_setup



import shutil
from pathlib import Path

import typer

app = typer.Typer()


@app.command()
def rename_and_move_for_nnunet(input_dir, output_dir, subdir_name=""):
    """
    Args:
        input_dir: absolute path to the directory with all the nifti cases
        output_dir: absolute path to the directory for nnunet inference
        subdir_name: optional name of subdirectory within case dir
    """
    output_dir = Path(output_dir)
    input_dir = Path(input_dir)
    output_dir.mkdir(exist_ok=True, parents=True)
    image_paths = input_dir.rglob("*.nii.gz")
    for image_path in image_paths:
        id_ = image_path.relative_to(input_dir).parts[0]
        if subdir_name in str(image_path):
            output_path = output_dir / (f"{id_}_0000.nii.gz")
            shutil.copy(str(image_path), str(output_path))



import datetime
import time
from pathlib import Path

import nibabel as nib
import numpy as np
import SimpleITK as sitk
from nilearn.image import resample_img, resample_to_img


def time_it(func):
    def wrapper(*args, **kwargs):
        start = time.time()
        result = func(*args, **kwargs)
        end = time.time()
        print(func.__name__ + " took " + str(end - start) + "sec")
        return result

    return wrapper


def convert_nrrd_to_nifti(nrrd_path, output_path):
    """
    Converts an image in NRRD format to NifTI format.
    """
    img = sitk.ReadImage(nrrd_path)
    sitk.WriteImage(img, output_path)


def resample_nifti(nifti_path, output_path, res=1.0):
    """
    Resamples a nifti to an isotropic resolution of res
    Args:
        nifti_path: path to a NifTI image
        output_path: path where to save the resampled NifTI image
    """
    assert Path(nifti_path).exists()
    nifti = nib.load(nifti_path)
    nifti_resampled = resample_img(
        nifti, target_affine=np.eye(3) * res, interpolation="nearest"
    )
    nib.save(nifti_resampled, output_path)


def resample_to_nifti(nifti_path, ref_path, output_path=None, interpolation="nearest"):
    """
    Resamples nifti to reference nifti, using nilearn resample_to_img, with paths as arguments.
    """
    if output_path is None:
        output_path = nifti_path
    nifti = nib.load(nifti_path)
    ref_nifti = nib.load(ref_path)
    nifti_resampled = resample_to_img(nifti, ref_nifti, interpolation)
    print(
        f"Mask size resamopled from {nifti.get_fdata().shape} to \
            {nifti_resampled.get_fdata().shape} [mask_path={output_path}]"
    )
    nib.save(nifti_resampled, output_path)


def combine_nifti_masks(mask1_path, mask2_path, output_path):
    """
    Args:
        mask1_path: abs path to the first nifti mask
        mask2_path: abs path to the second nifti mask
        output_path: abs path to saved concatenated mask
    """
    assert Path(mask1_path).exists
    assert Path(mask2_path).exists

    mask1 = nib.load(mask1_path)
    mask2 = nib.load(mask2_path)

    matrix1 = mask1.get_fdata()
    matrix2 = mask2.get_fdata()
    assert matrix1.shape == matrix2.shape

    new_matrix = np.zeros(matrix1.shape)
    new_matrix[matrix1 == 1] = 1
    new_matrix[matrix2 == 1] = 2
    new_matrix = new_matrix.astype(int)

    new_mask = nib.Nifti1Image(new_matrix, affine=mask1.affine, header=mask1.header)
    nib.save(new_mask, output_path)


def separate_nifti_masks(
    combined_mask_path, output_dir, label_dict=None, overwrite=False
):
    """
    Split multilabel nifti mask into separate binary nifti files.
    Args:
        combined_mask_path: abs path to the combined nifti mask
        output_path: abs path where to save separate masks
        label_dict: (optional) dictionary with names for each label
    """
    assert Path(combined_mask_path).exists()

    mask = nib.load(combined_mask_path)
    matrix = mask.get_fdata()

    labels = np.unique(matrix).astype(int)
    labels = labels[labels != 0]

    assert sorted(labels), sorted(list(label_dict.keys()))

    for label in labels:
        new_matrix = np.zeros(matrix.shape)
        new_matrix[matrix == label] = 1
        new_matrix = new_matrix.astype(int)

        new_mask = nib.Nifti1Image(new_matrix, affine=mask.affine, header=mask.header)

        if label_dict is not None:
            label_name = label_dict[label]
        else:
            label_name = label

        output_path = Path(output_dir) / f"seg_{label_name}.nii.gz"
        if (not output_path.exists()) or (overwrite is True):
            nib.save(new_mask, output_path)


def get_peak_from_histogram(bins, bin_edges):
    """
    Returns location of histogram peak.
    Can be applied on the output from np.histogram.
    In case of multiple equal peaks, returns locaion of the first one.

    Args :
        bins: values of histogram
        bin_edges: argument values at bin edges (len(bins)+1)

    Returns:
        peak_location: location of histogram peak (as a mean of mean edges)
    """
    assert len(bins) > 0
    assert len(bin_edges) == len(bins) + 1
    try:
        peak_bin = np.argmax(bins)
        print(peak_bin, "here i am!")
        peak_location = (bin_edges[peak_bin + 1] - bin_edges[peak_bin]) / 2

        return peak_location
    except Exception:
        raise ValueError("Error processing the bins.")


def calculate_age(dob):
    """
    Calculate the age of a person from his date of birth.
    """
    today = datetime.datetime.now()
    return today.year - dob.year - ((today.month, today.day) < (dob.month, dob.day))



import streamlit as st
import utils
from classrad.config import config
from classrad.data.dataset import Dataset
from classrad.models.classifier import MLClassifier
from classrad.training.trainer import Trainer


def show():
    """Shows the sidebar components for the template and returns
    user inputs as dict."""

    with st.sidebar:
        pass
    feature_df = utils.load_df("Choose a CSV file with radiomics features")
    st.write(feature_df)
    feature_df.dropna(axis="index", inplace=True)
    feature_df_colnames = feature_df.columns.tolist()
    col1, col2 = st.columns(2)
    with col1:
        label = st.selectbox("Select the label", feature_df_colnames)
    with col2:
        pat_id = st.selectbox("Select patient ID", feature_df_colnames)

    available_classifiers = config.AVAILABLE_CLASSIFIERS

    model_names = st.multiselect("Select the models", available_classifiers)
    num_features = st.slider(
        "Number of features", min_value=2, max_value=50, value=10
    )
    # Mlflow tracking
    # track_with_mlflow = st.checkbox("Track with mlflow?")

    # Model training
    start_training = st.button("Start training")
    if start_training:
        feature_colnames = [
            col
            for col in feature_df.columns
            if col.startswith(("original", "wavelet", "shape"))
        ]
        data = Dataset(feature_df, feature_colnames, label, "Harnstau")
        data.cross_validation_split_by_patient(patient_colname=pat_id)

        models = [MLClassifier(name) for name in model_names]
        result_dir = config.RESULT_DIR
        trainer = Trainer(
            dataset=data,
            models=models,
            result_dir=result_dir,
            num_features=num_features,
            meta_colnames=[pat_id, label],
        )
        with st.spinner("Training in progress..."):
            trainer.train_cross_validation()
        fig = trainer.dataset.boxplot_by_class()
        st.success(
            f"Training done! Predictions saved in your result directory \
            ({result_dir})"
        )
        st.plotly_chart(fig, use_container_width=True)


if __name__ == "__main__":
    show()



import itertools
import logging
import os
from collections import defaultdict
from os.path import join
from typing import Callable, List, Optional, Sequence, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
from radiomics import featureextractor
from SimpleITK import GetImageFromArray

logging.getLogger("radiomics").setLevel(
    logging.CRITICAL + 1
)  # pyradiomics makes a lot of noise


def equals_2(img):
    return img == 2


# taken from https://vincentblog.xyz/posts/medical-images-in-python-computed-tomography
def window_image(image, window_center, window_width):
    img_min = window_center - window_width // 2
    img_max = window_center + window_width // 2
    window_image = image.copy()
    window_image[window_image < img_min] = img_min
    window_image[window_image > img_max] = img_max

    return window_image


# taken from monai.transforms.utils
def is_positive(img):
    return img > 0


# adapted from monai.transforms.utils
def generate_spatial_bounding_box(
    img: np.ndarray,
    select_fn: Callable = is_positive,
    margin: Union[Sequence[int], int] = 0,
) -> Tuple[List[int], List[int]]:
    """
    generate the spatial bounding box of foreground in the image with start-end positions.
    Users can define arbitrary function to select expected foreground from the whole image or specified channels.
    And it can also add margin to every dim of the bounding box.
    The output format of the coordinates is:

        [1st_spatial_dim_start, 2nd_spatial_dim_start, ..., Nth_spatial_dim_start],
        [1st_spatial_dim_end, 2nd_spatial_dim_end, ..., Nth_spatial_dim_end]

    The bounding boxes edges are aligned with the input image edges.
    This function returns [-1, -1, ...], [-1, -1, ...] if there's no positive intensity.

    Args:
        img: source image to generate bounding box from.
        select_fn: function to select expected foreground, default is to select values > 0.
        margin: add margin value to spatial dims of the bounding box, if only 1 value provided, use it for all dims.
    """
    data = img
    data = np.any(select_fn(data), axis=0)
    ndim = len(data.shape)
    for m in margin:
        if m < 0:
            raise ValueError("margin value should not be negative number.")

    box_start = [0] * ndim
    box_end = [0] * ndim

    for di, ax in enumerate(itertools.combinations(reversed(range(ndim)), ndim - 1)):
        dt = data.any(axis=ax)
        if not np.any(dt):
            # if no foreground, return all zero bounding box coords
            return [0] * ndim, [0] * ndim

        min_d = max(np.argmax(dt) - margin[di], 0)
        max_d = max(
            data.shape[di] - max(np.argmax(dt[::-1]) - margin[di], 0), min_d + 1
        )
        box_start[di], box_end[di] = min_d, max_d

    return box_start, box_end


def create_maps(image_volume, seg_volume, save_dir, margin=50, num_displayed_slices=6):

    expanded_seg_volume = np.expand_dims(seg_volume, axis=0)
    coords_start, coords_end = generate_spatial_bounding_box(
        img=expanded_seg_volume, margin=[50, 50, 0]
    )
    print(
        f"Cropping the image of size {seg_volume.shape} to the region from {coords_start} to {coords_end}"
    )

    image_voi = image_volume[
        coords_start[0] : coords_end[0],
        coords_start[1] : coords_end[1],
        coords_start[2] : coords_end[2],
    ]  # needs rewriting into zip?
    seg_voi = seg_volume[
        coords_start[0] : coords_end[0],
        coords_start[1] : coords_end[1],
        coords_start[2] : coords_end[2],
    ]

    num_slices = image_voi.shape[2]
    selected_slices = [
        i * (num_slices - 1) // (num_displayed_slices + 1)
        for i in range(1, num_displayed_slices + 1)
    ]

    feature_classes = ["firstorder", "glcm", "glrlm", "glszm", "gldm", "ngtdm"]
    feature_count = {
        "firstorder": 19,
        "glcm": 24,
        "glrlm": 16,
        "glszm": 16,
        "gldm": 14,
        "ngtdm": 5,
    }
    extractor = featureextractor.RadiomicsFeatureExtractor()
    extractor.settings["normalize"] = False

    for feature_class in feature_classes:
        extractor.disableAllFeatures()
        extractor.enableFeatureClassByName(feature_class)
        num_features = feature_count[feature_class]

        fig, ax = plt.subplots(
            num_features + 1, num_displayed_slices, figsize=(20, 5 * (num_features))
        )
        fig.tight_layout(rect=[0, 0.03, 1, 0.95])
        fig.suptitle(f"{num_features} features from class {feature_class}", fontsize=25)
        print(f"Plotting maps for {num_features} radiomics features")

        for slice_cnt, slice_num in enumerate(selected_slices):
            image_slice = image_voi[:, :, slice_num]
            seg_slice = seg_voi[:, :, slice_num]
            image_slice = np.flip(image_slice.T, axis=0)
            seg_slice = np.flip(seg_slice.T, axis=0)
            xx, yy = np.meshgrid(
                np.arange(image_slice.shape[1]), np.arange(image_slice.shape[0])
            )
            region_labels = np.floor(xx / 8) * 64 + np.floor(yy / 8)
            prop_imgs = defaultdict(
                lambda: np.zeros_like(image_slice, dtype=np.float32)
            )
            out_df_list = []
            for patch_idx in np.unique(region_labels):
                # print(str(patch_idx), 'xdd')
                xx_box, yy_box = np.where(region_labels == patch_idx)
                c_block = image_slice[
                    xx_box.min() : xx_box.max(), yy_box.min() : yy_box.max()
                ]
                if c_block.shape[0] > 1 and c_block.shape[1] > 1:
                    c_block = np.expand_dims(c_block, axis=2)
                    c_block = c_block.astype(np.uint8)
                    c_label = np.ones_like(c_block).astype(np.uint8)
                    c_label[0, 0, 0] = 0
                    # print(c_block.shape, c_label.shape)
                    out_row = extractor.execute(
                        GetImageFromArray(c_block), GetImageFromArray(c_label)
                    )
                    for k, v in out_row.items():
                        if isinstance(v, (float, np.floating)) or isinstance(
                            v, (np.ndarray, np.generic)
                        ):
                            prop_imgs[k][region_labels == patch_idx] = v
                    out_df_list += [out_row]

            ax[0, slice_cnt].imshow(image_slice, cmap="gray")
            ax[0, slice_cnt].axis("off")
            ax[0, slice_cnt].set_title(f"Image - slice {slice_num}/{num_slices}")
            cnt = 1
            feature_names = list(prop_imgs.keys())
            feature_names = [
                name
                for name in feature_names
                if not "diagnostics" in name and not "shape" in name
            ]
            for feature_name in feature_names:
                map = prop_imgs[feature_name]
                maskedmap = np.ma.masked_array(map, mask=(seg_slice == 0))
                ax[cnt, slice_cnt].imshow(image_slice, cmap="gray")
                ax[cnt, slice_cnt].imshow(maskedmap, cmap="Spectral")
                ax[cnt, slice_cnt].axis("off")
                title = feature_name.split("_")[2]
                ax[cnt, slice_cnt].set_title(
                    f"{title} \n slice {slice_num}/{num_slices}"
                )
                cnt += 1

        os.makedirs(save_dir, exist_ok=True)
        save_path = join(save_dir, feature_class + "_maps.pdf")
        fig.savefig(save_path)



import streamlit as st
import utils
from classrad.evaluation.evaluator import Evaluator
from classrad.config import config


def show():
    """Shows the sidebar components for the template and returns
    user inputs as dict."""

    with st.sidebar:
        pass

    # LAYING OUT THE TOP SECTION OF THE APP
    col1, col2 = st.columns((2, 3))

    st.write(
        """
        #####
        Analyze the results with most widely used metrics such as
        AUC ROC curve, precision-recall curve and confusion matrix.
        """
    )
    result_df = utils.load_df("Choose a CSV file with predictions:")
    st.write(result_df)
    result_df_colnames = result_df.columns.tolist()
    label = st.selectbox("Select the label", result_df_colnames)
    # Evaluation
    evaluate = st.button("Evaluate!")
    if evaluate:
        evaluator = Evaluator(
            result_df=result_df,
            target=label,
            result_dir=config.RESULT_DIR,
        )
        evaluator.evaluate()
        st.write(evaluator.plot_roc_curve_all())
        st.write(evaluator.plot_confusion_matrix_all())
        st.write(
            f"""
            The best performing model in terms of AUC ROC in 5-fold
            cross-validation is ***{evaluator.best_model_name}**.
            This model is evaluated on the test set:
        """
        )
        st.write(evaluator.plot_test())


if __name__ == "__main__":
    show()



"""
Write a FeatureExtractor class on top of pyradiomics featureextractor,
which will extract features, given paths from a pandas df.
"""

import logging
import sys
from multiprocessing import Pool

import pandas as pd
from radiomics import featureextractor
from classrad.utils.utils import time_it
from tqdm import tqdm


class FeatureExtractor:
    """
    Class to extract features from a set of images.
    """

    def __init__(
        self,
        df,
        out_path,
        feature_set="pyradiomics",
        extraction_params=None,
        image_col="image_path",
        mask_col="mask_path",
        verbose=False,
        num_threads=None,
    ):
        """
        Initialize the Feature Extractor.
        :param df: pandas df with columns 'image_path' and 'mask_path'
        :param out_path: path to save the output csv
        :param extractor_params: path to the pyradiomics extractor params
        :param image_col: name of df column containing paths to images
        :param mask_col: name of df column containing paths to masks
        :param verbose: whether to print progress
        """
        self.df = df
        self.out_path = out_path
        self.feature_set = feature_set
        self.extraction_params = extraction_params
        self.image_col = image_col
        self.mask_col = mask_col
        self.verbose = verbose
        self.num_threads = num_threads
        self.feature_df = None
        self.extractor = None
        self.logger = logging.getLogger(__name__)
        self.logger.setLevel(logging.INFO)
        self.logger.addHandler(logging.StreamHandler(sys.stdout))
        self.logger.info("FeatureExtractor initialized")

    def extract_features(self, feature_set=None, verbose=None):
        """
        Extract features from a set of images.
        """
        if feature_set is not None:
            self.feature_set = feature_set
        if verbose is not None:
            self.verbose = verbose

        # Get the feature extractor
        self.logger.info("Initializing feature extractor")
        self.get_feature_extractor()

        self.logger.info("Initializing feature dataframe")
        self.initialize_feature_df()

        # Get the feature values
        self.logger.info("Extracting features")
        if self.num_threads is not None:
            self.get_features_parallel()
        else:
            self.get_features()
        self.save_feature_df()

    def get_feature_extractor(self):
        """
        Get the feature extractor.
        """
        if self.feature_set == "pyradiomics":
            self.extractor = featureextractor.RadiomicsFeatureExtractor(
                self.extraction_params
            )
        else:
            raise ValueError("Feature set not supported")
        return self

    def add_features_for_single_case(self, case):
        """
        Run extraction for one case and append results to feature_df
        Args:
            case: pd.Series describing a row of df
        """
        image_path = case[self.image_col]
        mask_path = case[self.mask_col]
        feature_vector = self.extractor.execute(image_path, mask_path)
        feature_series = pd.concat([case, pd.Series(feature_vector)])
        return feature_series

    def get_feature_names(self, case):
        image_path = case[self.image_col]
        mask_path = case[self.mask_col]
        feature_vector = self.extractor.execute(image_path, mask_path)
        feature_names = feature_vector.keys()
        return feature_names

    def initialize_feature_df(self):
        if self.feature_df is None:
            first_df_row = self.df.iloc[0]
            feature_names = self.get_feature_names(first_df_row)
            self.feature_df = self.df.copy()
            self.feature_df = self.feature_df.reindex(columns=feature_names)
        else:
            self.logger.info("Dataframe already has content!")
        return self

    @time_it
    def get_features(self):
        """
        Get the feature values.
        """
        rows = self.df.iterrows()
        for index, row in tqdm(rows):
            feature_series = self.add_features_for_single_case(row)
            self.feature_df = self.feature_df.append(feature_series, ignore_index=True)
        return self

    def save_feature_df(self):
        self.feature_df.to_csv(self.out_path, index=False)

    @time_it
    def get_features_parallel(self):
        try:
            _, df_rows = zip(*self.df.iterrows())
            p = Pool(self.num_threads)
            results = p.map(self.add_features_for_single_case, df_rows)
            self.feature_df = pd.concat(results, axis=1).T
        except Exception:
            print("Multiprocessing failed! :/")
        return self



from hypothesis.extra.pandas import data_frames, column, range_indexes
from hypothesis import given, settings, strategies as st
from classrad.data.dataset import Dataset
import pytest

simple_df = data_frames(
    columns=[
        column(name="ID", elements=st.integers()),
        column(name="Label", elements=st.booleans()),
        column(
            name="Feature1", elements=st.floats(min_value=0.0, max_value=1.0)
        ),
        column(
            name="Feature2",
            elements=st.floats(min_value=-1000.0, max_value=1000.0),
        ),
    ],
    index=range_indexes(min_size=100, max_size=100),
)


class TestDataset:
    # @classmethod
    # def setup_class(cls):
    #     pass

    # @classmethod
    # def teardown_class(cls):
    #     pass

    # @given(df=simple_df)
    # @settings(max_examples=1)
    # def setup_method(self):
    #     self.dataset = Dataset(
    #         dataframe=df,
    #         features=["Feature1"],
    #         target="Label",
    #         task_name="Testing",
    #     )

    # def teardown_method(self):
    #     del self.dataset

    @given(df=simple_df)
    @settings(max_examples=5)
    def test_init(self, df):
        dataset = Dataset(
            dataframe=df,
            features=["Feature1"],
            target="Label",
            task_name="Testing",
        )
        assert dataset.X.columns == ["Feature1"]
        assert dataset.y.name == "Label"

    @given(df=simple_df)
    @settings(max_examples=5)
    def test_split_dataset(self, df):
        dataset = Dataset(
            dataframe=df,
            features=["Feature1"],
            target="Label",
            task_name="Testing",
        )
        label_values = dataset.y.value_counts().tolist()
        label_values = sorted(label_values)
        if label_values[0] < 2:
            with pytest.raises(ValueError):
                dataset.split_dataset()
        dataset.split_dataset()
        assert dataset.X_train.shape[0] == 60
        assert dataset.X_val.shape[0] == 20
        assert dataset.X_test.shape[0] == 20
        assert dataset.y_train.shape[0] == 60
        assert dataset.y_val.shape[0] == 20
        assert dataset.y_test.shape[0] == 20

        assert (
            dataset.X_train.shape[1]
            == dataset.X_test.shape[1]
            == dataset.X_val.shape[1]
            == 1
        )

    @given(df=simple_df)
    @settings(max_examples=5)
    def test_cross_validation_split(self, df):
        dataset = Dataset(
            dataframe=df,
            features=["Feature1"],
            target="Label",
            task_name="Testing",
        )
        dataset.cross_validation_split()
        assert dataset.X_train.shape[0] == 80
        assert dataset.X_test.shape[0] == 20
        assert dataset.y_train.shape[0] == 80
        assert dataset.y_test.shape[0] == 20



def get_subplots_dimensions(n_plots):
    """
    For given number of plots returns 'optimal' rows x columns distribution of subplots and figure size.
    Args:
        n_plots [int] - number of subplots to be includeed in the plot
    Returns:
        nrows [int] - suggested number of rows
        ncols [int] - suggested number of columns
        figsize [tuple[int, int]] - suggested figsize
    """
    if n_plots == 1:
        nrows = 1
        ncols = 1
        figsize = (12, 7)
    elif n_plots == 2:
        nrows = 1
        ncols = 2
        figsize = (13, 6)
    elif n_plots == 3:
        nrows = 1
        ncols = 3
        figsize = (20, 5)
    elif n_plots == 4:
        nrows = 2
        ncols = 2
        figsize = (14, 8)
    elif n_plots in [5, 6]:
        nrows = 2
        ncols = 3
        figsize = (20, 9)
    elif n_plots == 9:
        nrows = 3
        ncols = 3
        figsize = (18, 12)
    elif n_plots == 10:
        nrows = 2
        ncols = 5
        figsize = (20, 7)
    elif n_plots > 4:
        nrows = n_plots // 4 + 1
        ncols = 4
        figsize = (20, 7 + 5 * nrows)

    return nrows, ncols, figsize


# Review the following functions for usefulness
# def plot_for_all(func):
#     def wrapper(model_names, predictions, *args, **kwargs):
#         nrows, ncols, figsize = get_subplots_dimensions(len(models))
#         fig, axs = plt.subplots(nrows=nrows, ncols=ncols, figsize=figsize)
#         for i, model_predictions in enumerate(predictions):
#             ax = fig.axes[i]
#             model_name = model_names[i]
#             func(model_name, model_predictions, ax=ax)
#         plt.tight_layout()
#         plt.show()

#     return wrapper



import os
from pathlib import Path

import typer
from nipype.interfaces.dcm2nii import Dcm2niix

app = typer.Typer()


def get_dcm2niix_converter(dicom_dir: Path, save_dir: Path):
    """
    Args:
        dicom_dir: directory with DICOM Study
        save_dir: directory where to save the nifti
    Returns:
        converter: nipype Dcm2niix object based on rodenlab dcm2niix with basic
                   default parameters to merge Dicom series into 3d Volumes
    """
    converter = Dcm2niix()
    converter.inputs.source_dir = str(dicom_dir)
    converter.inputs.output_dir = str(save_dir)
    converter.inputs.merge_imgs = True

    return converter


@app.command()
def dicom_to_nifti(input_dir: str, output_dir: str, subdir_name: str = ""):
    """
    Args:
        input_dir: absolute path to the directory with all the cases containing
                dicoms
        output_dir: absolute path to the directory where to save nifties
        subdir_name: optional name of subdirectory within case dir
    """
    input_dir = Path(input_dir)
    output_dir = Path(output_dir)
    for id_ in os.listdir(input_dir):
        dicom_dir = input_dir / id_ / subdir_name
        save_dir = output_dir / id_ / subdir_name
        save_dir.mkdir(exist_ok=True)
        assert dicom_dir.exists()

        converter = get_dcm2niix_converter(dicom_dir, save_dir)
        converter.run()



from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.feature_selection import SelectKBest, f_classif


class FeatureSelector(BaseEstimator, TransformerMixin):
    def __init__(self, method, k):
        self.method = method
        self.k = k

    def fit(self, X, y):
        if self.method == "Select K Best":
            self.selector = SelectKBest(f_classif, k=self.k)
        else:
            raise ValueError("Invalid method")
        self.selector.fit(X, y)
        return self

    def transform(self, X):
        return self.selector.transform(X)

    def inverse_transform(self, X):
        return self.selector.inverse_transform(X)



import os
from os.path import isdir, join, splitext

if __name__ == '__main__':

    image_dir =
    segmentation_dir =

    all_images = os.listdir(image_dir)
    nii_images = [fname in all_images if fname.endswith('.nii.gz')]

    all_segmentations = os.listdir(seg_dir)
    nii_segmentations = [fname in all_segmentations if fname.endswith('.nii.gz')]

    if len(nii_images) == 0:
        print('No Nifti images found in the directory!')
    elif len(nii_images) != len(nii_segmentations):
        print('Number of images and segmentations not matching!')
    else:
        for img_id in nii_images:
            if not endswith('_img.nii.gz'):
                print('False name of the image file:', img_id)
            else:
                img_path = join(image_dir, img_id)
                seg_id = img_id[:-10] + 'seg.nii.gz'
                seg_path = join(seg_dir, seg_id)
                if not exists(seg_path):
                    print('Image ', img_id, 'has no corresponding segmentation')
                else:
                    img = nib.load(img_path)
                    seg = nib.load(seg_path)



import numpy as np
from imblearn.metrics import sensitivity_specificity_support
from scipy import stats
from sklearn.metrics import confusion_matrix, roc_auc_score
from sklearn.utils import resample


def wilcoxon_unpaired(x, y, alternative="two-sided"):
    """
    Test for differences between unpaired samples.
    """
    stat, p = stats.ranksums(x, y, alternative=alternative)
    if p < 0.001:
        p_string = "< 0.001"
    else:
        p_string = str(np.round(p, 3))
    return stat, p_string


def get_sens_spec(y_true, y_pred):
    """
    Args:
        y_true: list of binary ground-truth labels
        y_pred: list of binary predictions
    Returns:
        sensitivity: True Positive Rate = Sensitivity
        specificity: True Negative Rate = Specificity (TNR = 1-FPR)
    """
    stat_table = sensitivity_specificity_support(y_true, y_pred)
    sensitivity = stat_table[1][0]
    specificity = stat_table[1][1]
    return sensitivity, specificity


def describe_auc(y_true, y_pred_proba):
    """
    Get mean AUC and 95% Confidence Interval from bootstrapping.
    """
    AUCs = []
    num_folds = 1000
    for i in range(num_folds):
        boot_y_pred, boot_y_true = resample(
            y_pred_proba,
            y_true,
            replace=True,
            n_samples=len(y_true),
            random_state=i + 10,
        )
        AUC = roc_auc_score(boot_y_true, boot_y_pred)
        AUCs.append(AUC)

    mean_AUC = np.mean(AUCs)
    lower_CI_bound = np.percentile(AUCs, 2.5)
    upper_CI_bound = np.percentile(AUCs, 97.5)
    print(f"Results from {num_folds} folds of bootrapping:")
    print(f"Mean AUC={mean_AUC}, 95% CI [{lower_CI_bound}, {upper_CI_bound}]")

    return mean_AUC, lower_CI_bound, upper_CI_bound


def describe_stat(y_true, y_pred):
    """
    Get sensitivity and specificity with corresponding CIs using bootstrapping.
    Args:
        y_true: list of binary ground-truth labels
        y_pred:
    """
    sens, spec = get_sens_spec(y_true, y_pred)
    cm = confusion_matrix(y_true, y_pred)

    print(sens, spec)
    print(cm)

    sens_all, spec_all = [], []
    num_folds = 1000
    for i in range(num_folds):
        boot_y_pred, boot_y_true = resample(
            y_pred, y_true, replace=True, n_samples=len(y_true), random_state=i
        )
        sens, spec = get_sens_spec(boot_y_true, boot_y_pred)
        sens_all.append(sens)
        spec_all.append(spec)
    sens_all, spec_all = np.array(sens_all), np.array(spec_all)

    mean_sens = np.mean(sens_all)
    lower_CI_bound_sens = np.percentile(sens_all, 2.5)
    upper_CI_bound_sens = np.percentile(sens_all, 97.5)

    mean_spec = np.mean(spec_all)
    lower_CI_bound_spec = np.percentile(spec_all, 2.5)
    upper_CI_bound_spec = np.percentile(spec_all, 97.5)

    print(f"Results from {num_folds} folds of bootrapping:")
    print(
        f"Sensitivity: Mean={mean_sens}, 95% CI [{lower_CI_bound_sens}, \
          {upper_CI_bound_sens}]"
    )
    print(
        f"Specificity: Mean={mean_spec}, 95% CI [{lower_CI_bound_spec}, \
          {upper_CI_bound_spec}]"
    )

    return (mean_sens, lower_CI_bound_sens, upper_CI_bound_sens), (
        mean_spec,
        lower_CI_bound_spec,
        upper_CI_bound_spec,
    )


# Review, not a particularly nice implementation
def roc_auc_ci(y_true, y_pred_proba, positive=1):
    """
    Get 95% Confidence Interval (CI) for AUC, assuming normal distribution.
    Args:
        y_true: list of binary ground-truth labels
        y_pred_proba: list of probabilities of positive class
        positive: class considered as positive
    Returns:
        AUC: AUC score
        lower: lower bound of 95% CI
        upper: upper bound of 95% CI
    """
    AUC = roc_auc_score(y_true, y_pred_proba)
    N1 = np.sum(y_true == positive)
    N2 = np.sum(y_true != positive)
    Q1 = AUC / (2 - AUC)
    Q2 = 2 * AUC ** 2 / (1 + AUC)
    SE_AUC = np.sqrt(
        (AUC * (1 - AUC) + (N1 - 1) * (Q1 - AUC ** 2) + (N2 - 1) * (Q2 - AUC ** 2))
        / (N1 * N2)
    )
    lower = AUC - 1.96 * SE_AUC
    upper = AUC + 1.96 * SE_AUC
    if lower < 0:
        lower = 0
    if upper > 1:
        upper = 1

    return AUC, lower, upper



import streamlit as st


def show():
    with st.sidebar:
        modalities = ["CT", "MRI"]
        st.selectbox("Modality", modalities)
        organs = [
            "Brain",
            "Lung",
            "Heart",
            "Liver",
            "Kidney",
            "Spleen",
            "Pancreas",
            "Intestine",
        ]
        st.selectbox("Organ", organs)


if __name__ == "__main__":
    show()



import numpy as np
from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.ensemble import AdaBoostClassifier, RandomForestClassifier
from sklearn.gaussian_process import GaussianProcessClassifier
from sklearn.linear_model import LogisticRegression
from sklearn.svm import SVC
from xgboost import XGBClassifier


class MLClassifier(ClassifierMixin):
    def __init__(self, classifier_name, classifier_parameters={}):
        self.classifier = None
        self.classifier_name = classifier_name
        self.classifier_parameters = classifier_parameters
        self.available_classifiers = [
            "Random Forest",
            "AdaBoost",
            "Logistic Regression",
            "SVM",
            "Gaussian Process Classifier",
            "XGBoost",
        ]
        self.select_classifier()

    def select_classifier(self):
        if self.classifier_name == "Random Forest":
            self.classifier = RandomForestClassifier(**self.classifier_parameters)
        elif self.classifier_name == "AdaBoost":
            self.classifier = AdaBoostClassifier(**self.classifier_parameters)
        elif self.classifier_name == "Logistic Regression":
            self.classifier = LogisticRegression(
                max_iter=1000, **self.classifier_parameters
            )
        elif self.classifier_name == "SVM":
            self.classifier = SVC(probability=True, **self.classifier_parameters)
        elif self.classifier_name == "Gaussian Process Classifier":
            self.classifier = GaussianProcessClassifier(**self.classifier_parameters)
        elif self.classifier_name == "XGBoost":
            self.classifier = XGBClassifier(
                verbosity=0,
                silent=True,
                use_label_encoder=False,
                **self.classifier_parameters,
            )
        else:
            raise ValueError("Classifier name not recognized")

    def fit(self, X, y):
        if self.classifier is None:
            raise AssertionError("Run .select_classifier first!")
        else:
            self.classifier.fit(X, y)

    def predict(self, X):
        return self.classifier.predict(X)

    def predict_proba(self, X):
        return self.classifier.predict_proba(X)

    def predict_label_and_proba(self, X):
        y_pred = self.classifier.predict(X)
        y_pred_proba = self.classifier.predict_proba(X)[:, 1]

        return y_pred, y_pred_proba

    def score(self, X, y):
        return self.classifier.score(X, y)

    def get_params(self, deep):
        return self.classifier.get_params(deep)

    def set_params(self, params):
        self.classifier.set_params(**params)
        self.classifier_parameters = params
        return self

    def get_available_classifiers(self):
        return self.available_classifiers

    def update_classifier(self, new_classifier_name):
        self.classifier_name = new_classifier_name

    def feature_importance(self):
        if self.classifier_name == "Logistic Regression":
            importance = self.classifier.coef_[0]
        elif self.classifier_name in ["AdaBoost", "Random Forest", "XGBoost"]:
            importance = self.classifier.feature_importances_
        else:
            raise ValueError(
                f"For classifier {self.classifier_name} feature \
                               importance could not be calculated."
            )
        return importance


class EnsembleClassifier(BaseEstimator, ClassifierMixin):
    """
    Ensemble classifier for MLClassifiers.

    Parameters:
    clf : `iterable`
      A list of classifier objects.
    weights : `list` (default: `None`)
      If `None`, the majority rule voting will be applied to the predicted class labels.
        If a list of weights (`float` or `int`) is provided, the averaged raw probabilities (via `predict_proba`)
        will be used to determine the most confident class label.
    """

    def __init__(self, classifier_list, weights=None):
        self.classifier_list = classifier_list
        self.weights = weights

    def fit(self, X, y):
        """
        Fit the scikit-learn estimators.

        Parameters:
        X : numpy array, shape = [n_samples, n_features]
            Training data
        y : list or numpy array, shape = [n_samples]
            Class labels
        """
        for clf in self.classifier_list:
            clf.fit(X, y)

    def predict(self, X):
        """
        Parameters:
        X : numpy array, shape = [n_samples, n_features]

        Returns:
        maj : list or numpy array, shape = [n_samples]
            Predicted class labels by majority rule
        """
        self.classes_ = np.asarray([clf.predict(X) for clf in self.classifier_list])
        # if self.weights:
        #    avg = self.predict_proba(X)
        #         maj = np.apply_along_axis(lambda x: max(enumerate(x), key=operator.itemgetter(1))[0], axis=1, arr=avg)
        # else:
        #     maj = np.asarray(
        #         [
        #             np.argmax(np.bincount(self.classes_[:, c]))
        #             for c in range(self.classes_.shape[1])
        #         ]
        #     )

        # return maj

    def predict_proba(self, X):
        """
        Parameters:
        X : numpy array, shape = [n_samples, n_features]

        Returns:
        avg : list or numpy array, shape = [n_samples, n_probabilities]
            Weighted average probability for each class per sample.
        """
        self.probas_ = [clf.predict_proba(X) for clf in self.classifier_list]
        avg = np.average(self.probas_, axis=0, weights=self.weights)

        return avg

    def get_classifier_list(self):
        return self.classifier_list



import argparse
import os
from os.path import exists, isdir, join

import nibabel as nib
from utils import create_maps

if __name__ == "__main__":

    parser = argparse.ArgumentParser(
        description="Generate radiomics maps from image and ROI"
    )
    parser.add_argument(
        "-i",
        "--input",
        help="parent folder containing *images* and *segmentations* folders",
    )
    parser.add_argument("-o", "--output", help="folder to save the generated maps")
    parser.add_argument(
        "-m", "--margin", help="margin in pixel for the map size around ROI"
    )
    args = parser.parse_args()

    input_dir = args.input
    if input_dir is None:
        input_dir = "/data"
    assert isdir(input_dir)

    output_dir = args.output
    if output_dir is None:
        output_dir = join(input_dir, "results")
    image_dir = join(input_dir, "images")
    segmentation_dir = join(input_dir, "segmentations")
    margin = args.margin
    if margin is None:
        margin = 50
    else:
        margin = int(margin)

    all_images = os.listdir(image_dir)
    nii_images = [fname for fname in all_images if fname.endswith(".nii.gz")]

    all_segmentations = os.listdir(segmentation_dir)
    nii_segmentations = [
        fname for fname in all_segmentations if fname.endswith(".nii.gz")
    ]

    if len(nii_images) == 0:
        print("No Nifti images found in the directory!")
    elif len(nii_images) != len(nii_segmentations):
        print("Number of images and segmentations not matching!")
    else:
        for img_id in nii_images:
            if not img_id.endswith("_img.nii.gz"):
                print("False name of the image file:", img_id)
            else:
                img_path = join(image_dir, img_id)
                seg_id = img_id[:-11] + "_seg.nii.gz"
                seg_path = join(segmentation_dir, seg_id)
                if not exists(seg_path):
                    print("Image ", img_id, "has no corresponding segmentation")
                else:
                    img = nib.load(img_path).get_fdata().astype(int)
                    seg = nib.load(seg_path).get_fdata().astype(int)
                    save_dir = join(output_dir, img_id[:-11])
                    create_maps(img, seg, save_dir, margin)



"""
Create an Evaluator class to evaluate predictions for classification task
in terms of ROC AUC and sensitivity/specificity.
The models to evaluate are created on top of sklearn classifiers.
"""
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from classrad.utils.visualization import get_subplots_dimensions
from sklearn.metrics import auc, confusion_matrix, precision_recall_curve, roc_curve

from .metrics import roc_auc_score
from .utils import (
    common_roc_settings,
    get_fpr_tpr_auc,
    get_sensitivity_specificity,
    get_youden_threshold,
)


class Evaluator:
    def __init__(
        self,
        result_df,
        target,
        result_dir,
        n_jobs=1,
        random_state=None,
    ):
        self.result_df = result_df
        self.train_results = self.result_df[self.result_df["test"] == 0]
        self.test_results = self.result_df[self.result_df["test"] == 1]
        self.fold_results = self.train_results.groupby("cv_split").agg(pd.Series.tolist)
        self.target = target
        self.test_labels = self.test_results[self.target].tolist()
        self.train_labels = self.train_results[self.target].tolist()
        self.result_dir = result_dir
        self.n_jobs = n_jobs
        self.random_state = random_state
        self.scores = None
        self.predictions = None
        self.predictions_proba = None
        self.model_names = None
        self.best_model_name = None
        self.best_model_idx = None
        self.best_model_score_test = None
        self.best_model_threshold = None

        self.result_dir.mkdir(parents=True, exist_ok=True)

    def update_model_names(self):
        colnames = self.result_df.columns.tolist()
        relevant_colnames = [name for name in colnames if name.endswith("pred")]
        self.model_names = [colname.split("_")[0] for colname in relevant_colnames]
        return self

    def update_predictions(self):
        self.predictions, self.predictions_proba = {}, {}
        for model_name in self.model_names:
            self.predictions[model_name] = self.train_results[f"{model_name}_pred"]
            self.predictions_proba[model_name] = self.train_results[
                f"{model_name}_pred_proba"
            ]
        return self

    def evaluate(self):
        """
        Evaluate the models.
        """
        self.update_model_names()
        self.scores = {"cv": {}, "test": {}}
        for model_name in self.model_names:
            aucs = []
            for fold in range(len(self.fold_results)):
                fold_labels = self.fold_results[self.target][fold]
                fold_preds_proba = self.fold_results[f"{model_name}_pred_proba"][fold]
                aucs.append(roc_auc_score(fold_labels, fold_preds_proba))

            model_mean_score = np.round(np.mean(aucs), 3)
            model_std_score = np.round(np.std(aucs), 3)
            self.scores["cv"][model_name] = (model_mean_score, model_std_score)
            print(
                f"For {model_name} in 5-fold CV AUC = {model_mean_score} +/- {model_std_score}"
            )
        self.update_predictions()
        self.update_best_model()
        print(
            f"Best model: {self.best_model_name} - AUC on test set = {self.best_model_score_test}"
        )

        return self

    def get_roc_threshold(self):
        y_true = self.train_labels
        y_pred_proba = self.predictions_proba[self.best_model_name]
        _, _, self.best_model_threshold = get_youden_threshold(y_true, y_pred_proba)
        return self

    def update_best_model(self):
        self.best_model_idx = np.argmax([t[0] for t in self.scores["cv"].values()])
        self.best_model_name = self.model_names[self.best_model_idx]
        self.predictions_proba_test = self.test_results[
            f"{self.best_model_name}_pred_proba"
        ]
        self.best_model_score_test = roc_auc_score(
            self.test_labels, self.predictions_proba_test
        )
        self.get_roc_threshold()
        return self

    def plot_roc_curve_cross_validation(self, model_name, ax, title=None):
        """
        Plot the ROC curve.
        """
        y_true = self.train_labels
        y_pred_proba = self.predictions_proba[model_name]
        (auc_mean, auc_std) = self.scores["cv"][model_name]
        fpr, tpr, roc_auc = get_fpr_tpr_auc(y_true, y_pred_proba)
        label = f"Cumulative AUC={roc_auc}, mean AUC={auc_mean}+/-{auc_std}"
        ax.plot(fpr, tpr, lw=3, alpha=0.8, label=label)
        # ax.legend(label, fontsize=10)
        if title:
            ax.set_title(title)
        else:
            ax.set_title(model_name)
        common_roc_settings(ax)

        return ax

    def plot_optimal_point_test(self, y_true, y_pred_proba, ax):
        thr = self.best_model_threshold
        y_pred = y_pred_proba > thr
        fpr, tpr, _ = roc_curve(y_true, y_pred)
        fpr_point = fpr[1]
        tpr_point = tpr[1]
        sens, spec = get_sensitivity_specificity(y_true, y_pred, thr)
        point_label = f"Sensitivity = {sens}, Specificity = {spec}"
        ax.plot(
            [fpr_point],
            [tpr_point],
            marker="h",
            markersize=10,
            color="black",
            label=point_label,
        )
        return self

    def plot_roc_curve_test(self, model_name, ax, title=None):
        """
        Plot the ROC curve.
        """
        y_true = self.test_labels
        y_pred_proba = self.test_results[f"{model_name}_pred_proba"]
        fpr, tpr, roc_auc = get_fpr_tpr_auc(y_true, y_pred_proba)
        label = f"{model_name} - AUC = {roc_auc}"
        ax.plot(fpr, tpr, lw=3, alpha=0.8, label=label)
        self.plot_optimal_point_test(y_true, y_pred_proba, ax)
        if title:
            ax.set_title(title)
        else:
            ax.set_title("ROC Curve")
        common_roc_settings(ax)

        return self

    def plot_roc_curve_all(self, title=None):
        """
        Plot the ROC Curve for all models.
        """
        nrows, ncols, figsize = get_subplots_dimensions(len(self.model_names))
        fig, axs = plt.subplots(nrows=nrows, ncols=ncols, figsize=figsize)
        for i, model_name in enumerate(self.model_names):
            ax = fig.axes[i]
            fig.axes[i] = self.plot_roc_curve_cross_validation(model_name, ax=ax)
        if title:
            fig.suptitle(title)
        else:
            fig.suptitle(f"ROC Curve for training set in 5-fold cross-validation.")
        fig.tight_layout()
        plt.show()
        fig.savefig(self.result_dir / "ROC.png", bbox_inches="tight", dpi=fig.dpi)

        return fig

    def plot_precision_recall_curve_test(self, model_name, ax=None, title=None):
        """
        Plot the precision recall curve.
        """
        y_true = self.test_labels
        y_pred_proba = self.test_results[f"{model_name}_pred_proba"]
        precision, recall, thresholds = precision_recall_curve(y_true, y_pred_proba)
        auc_score = np.round(auc(recall, precision), 3)
        ax.plot(
            recall, precision, lw=2, alpha=0.8, label=f"{model_name} AUC={auc_score}"
        )
        ax.set_xlabel("Recall")
        ax.set_ylabel("Precision")
        ax.legend(loc="lower left")
        if title:
            ax.set_title(title)
        else:
            ax.set_title("Precision-Recall Curve")

        return self

    def plot_precision_recall_curve_all(self, title=None):
        """
        Plot the precision recall curve for all models.
        """
        n_models = len(self.model_names)
        nrows, ncols, figsize = get_subplots_dimensions(n_models)
        fig, axs = plt.subplots(nrows=nrows, ncols=ncols, figsize=figsize)
        for i in range(n_models):
            model_name = self.model_names[i]
            ax = fig.axes[i]
            self.plot_precision_recall_curve(model_name, ax=ax)
        if title:
            fig.suptitle(title)
        else:
            fig.suptitle(f"Precision-Recall Curve for test dataset")
        fig.tight_layout()
        plt.show()

        return fig

    def plot_confusion_matrix_cross_validation(self, model_name, ax=None):
        """
        Plot the confusion matrix for a single model.
        """
        y_true = self.train_labels
        y_pred = self.predictions[model_name]
        cm = confusion_matrix(y_true, y_pred)
        sns.heatmap(cm, annot=True, fmt="d", ax=ax)
        ax.set_xlabel("Predicted")
        ax.set_ylabel("Actual")
        ax.set_title(model_name)

        return self

    def plot_confusion_matrix_test(self, model_name, ax=None, title=None):
        """
        Plot the confusion matrix for a single model.
        """
        y_true = self.test_labels
        y_pred = self.test_results[f"{model_name}_pred"]
        cm = confusion_matrix(y_true, y_pred)
        sns.heatmap(cm, annot=True, fmt="d", ax=ax)
        ax.set_xlabel("Predicted")
        ax.set_ylabel("Actual")
        if title:
            ax.set_title(title)
        else:
            ax.set_title("Confusion Matrix")

        return self

    def plot_confusion_matrix_all(self, title=None):
        """
        Plot the confusion matrix for all models.
        """
        nrows, ncols, figsize = get_subplots_dimensions(len(self.model_names))
        fig, axs = plt.subplots(nrows=nrows, ncols=ncols, figsize=figsize)
        for i, model_name in enumerate(self.model_names):
            ax = fig.axes[i]
            self.plot_confusion_matrix_cross_validation(model_name, ax=ax)
        if title:
            fig.suptitle(title)
        else:
            fig.suptitle(f"Confusion Matrix for 5-fold cross-validation")
        plt.tight_layout()
        plt.show()
        fig.savefig(self.result_dir / "confusion_matrix.png", bbox_inches="tight")

        return fig

    def plot_test(self, title=None):
        model_name = self.best_model_name

        nrows, ncols, figsize = get_subplots_dimensions(3)
        fig, axs = plt.subplots(nrows=nrows, ncols=ncols, figsize=figsize)

        self.plot_roc_curve_test(model_name, ax=axs[0])
        self.plot_precision_recall_curve_test(model_name, ax=axs[1])
        self.plot_confusion_matrix_test(model_name, ax=axs[2])
        if title:
            fig.suptitle(title)
        else:
            fig.suptitle(f"Results on test dataset using {model_name}")
        fig.tight_layout()
        plt.show()
        fig.savefig(self.result_dir / "test.png", bbox_inches="tight")

        return fig

    def plot_all_cross_validation(self):
        """
        Plot all the graphs on the cross-validation dataset.
        """
        self.plot_roc_curve_all()
        # self.plot_precision_recall_curve_all()
        self.plot_confusion_matrix_all()

        return self



def evaluate_folder(
    folder_with_gts: str, folder_with_predictions: str, labels: tuple, **metric_kwargs
):
    """
    writes a summary.json to folder_with_predictions
    :param folder_with_gts: folder where the ground truth segmentations are saved. Must be nifti files.
    :param folder_with_predictions: folder where the predicted segmentations are saved. Must be nifti files.
    :param labels: tuple of int with the labels in the dataset. For example (0, 1, 2, 3) for Task001_BrainTumour.
    :return:
    """
    files_gt = subfiles(folder_with_gts, suffix=".nii.gz", join=False)
    files_pred = subfiles(folder_with_predictions, suffix=".nii.gz", join=False)
    assert all(
        [i in files_pred for i in files_gt]
    ), "files missing in folder_with_predictions"
    assert all([i in files_gt for i in files_pred]), "files missing in folder_with_gts"
    test_ref_pairs = [
        (join(folder_with_predictions, i), join(folder_with_gts, i)) for i in files_pred
    ]
    res = aggregate_scores(
        test_ref_pairs,
        json_output_file=join(folder_with_predictions, "summary.json"),
        num_threads=8,
        labels=labels,
        **metric_kwargs
    )
    return res


def nnunet_evaluate_folder():
    import argparse

    parser = argparse.ArgumentParser(
        "Evaluates the segmentations located in the folder pred. Output of this script is "
        "a json file. At the very bottom of the json file is going to be a 'mean' "
        "entry with averages metrics across all cases"
    )
    parser.add_argument(
        "-ref",
        required=True,
        type=str,
        help="Folder containing the reference segmentations in nifti " "format.",
    )
    parser.add_argument(
        "-pred",
        required=True,
        type=str,
        help="Folder containing the predicted segmentations in nifti "
        "format. File names must match between the folders!",
    )
    parser.add_argument(
        "-l",
        nargs="+",
        type=int,
        required=True,
        help="List of label IDs (integer values) that should "
        "be evaluated. Best practice is to use all int "
        "values present in the dataset, so for example "
        "for LiTS the labels are 0: background, 1: "
        "liver, 2: tumor. So this argument "
        "should be -l 1 2. You can if you want also "
        "evaluate the background label (0) but in "
        "this case that would not gie any useful "
        "information.",
    )
    args = parser.parse_args()
    return evaluate_folder(args.ref, args.pred, args.l)



"""
Create a dataloader class from a dataframe, load selected columns as X and a
column as Y. Add function to split into training, validation and test sets or
stratified split or cross-validation split.
"""
import numpy as np
import pandas as pd
from typing import List
import plotly.graph_objects as go
from plotly.subplots import make_subplots
from classrad.utils.statistics import wilcoxon_unpaired
from classrad.utils.visualization import get_subplots_dimensions
from sklearn.feature_selection import SelectKBest, f_classif
from sklearn.model_selection import (
    StratifiedGroupKFold,
    StratifiedKFold,
    train_test_split,
)
from sklearn.preprocessing import MinMaxScaler

from classrad.config import config


class Dataset:
    """
    Store the data and labels, split into training/test sets, select features and show them.
    """

    def __init__(
        self,
        dataframe: pd.DataFrame,
        features: List[str],
        target: str,
        task_name: str = "",
        random_state: int = config.SEED,
    ):
        self.df = dataframe
        self.features = features
        self.target = target
        self.task_name = task_name
        self.random_state = random_state
        self.X = self.df[self.features]
        self.y = self.df[self.target]
        self.X_train = None
        self.X_val = None
        self.X_test = None
        self.y_train = None
        self.y_val = None
        self.y_test = None
        self.X_train_fold = None
        self.X_val_fold = None
        self.y_train_fold = None
        self.y_val_fold = None
        self.labels_cv_folds = []
        self.cv_split_generator = None
        self.cv_splits = None
        self.best_features = None
        self.scaler = MinMaxScaler()
        self.feature_selector = None

    def split_dataset(self, test_size: float = 0.2, val_size: float = 0.2):
        """
        Performs stratified split into 3 subsets:
            - test (`test_size` cases),
            - validation (`val_size` cases), and
            - training (rest).
        """
        self.X_train, self.X_test, self.y_train, self.y_test = train_test_split(
            self.X,
            self.y,
            test_size=test_size,
            stratify=self.y,
            random_state=self.random_state,
        )
        val_proportion_of_train = val_size / (1 - test_size)
        self.X_train, self.X_val, self.y_train, self.y_val = train_test_split(
            self.X_train,
            self.y_train,
            test_size=val_proportion_of_train,
            stratify=self.y_train,
            random_state=self.random_state,
        )
        return self

    def create_cross_validation_labels(self):
        """
        ?
        """
        for _, (train_idx, val_idx) in enumerate(self.cv_splits):
            self.get_cross_validation_fold(train_idx, val_idx)
            self.labels_cv_folds.extend(self.y_val_fold.values)
        return self

    def cross_validation_split(self, n_splits: int = 5, test_size: float = 0.2):
        """
        Perform stratified split into:
            - test (`test_size` cases),
            - training:
                - split info `n_splits` folds with stratified k-fold cross-validation.
        """
        self.X_train, self.X_test, self.y_train, self.y_test = train_test_split(
            self.X,
            self.y,
            test_size=test_size,
            stratify=self.y,
            random_state=self.random_state,
        )
        kf = StratifiedKFold(
            n_splits=n_splits, shuffle=True, random_state=self.random_state
        )
        self.cv_split_generator = kf.split(self.X_train, self.y_train)
        self.cv_splits = list(self.cv_split_generator)
        self.create_cross_validation_labels()
        return self

    def cross_validation_split_by_patient(
        self, patient_colname, n_splits=5, test_size=0.2
    ):
        train_inds, test_inds = next(
            StratifiedGroupKFold(
                n_splits=int(np.round(1 / test_size)),
                shuffle=True,
                random_state=self.random_state,
            ).split(self.df, self.y, groups=self.df[patient_colname])
        )
        self.X_train = self.X.iloc[train_inds]
        self.X_test = self.X.iloc[test_inds]
        self.y_train = self.y.iloc[train_inds]
        self.y_test = self.y.iloc[test_inds]
        df_train = self.df.iloc[train_inds]
        kf = StratifiedGroupKFold(
            n_splits=n_splits, shuffle=True, random_state=self.random_state
        )
        self.cv_split_generator = kf.split(
            self.X_train, self.y_train, groups=df_train[patient_colname]
        )
        self.cv_splits = list(self.cv_split_generator)
        self.create_cross_validation_labels()
        return self

    def get_cross_validation_fold(self, train_index, val_index):
        if self.cv_splits is None:
            print("No folds found. Try running 'cross_validation_split' first.")
        else:
            self.X_train_fold, self.X_val_fold = (
                self.X_train.iloc[train_index],
                self.X_train.iloc[val_index],
            )
            self.y_train_fold, self.y_val_fold = (
                self.y_train.iloc[train_index],
                self.y_train.iloc[val_index],
            )
        return self

    def split_train_test_from_column(self, column_name, test_value):
        self.df_test = self.df[self.df[column_name] == test_value]
        self.df_train = self.df[self.df[column_name] != test_value]
        self.X_test = self.df_test[self.features]
        self.y_test = self.df_test[self.target]
        self.X_train = self.df_train[self.features]
        self.y_train = self.df_train[self.target]
        return self

    def split_dataset_test_from_column(self, column_name, test_value, val_size=0.2):
        self.split_train_test_from_column(column_name, test_value)
        self.X_train, self.X_val, self.y_train, self.y_val = train_test_split(
            self.X_train,
            self.y_train,
            test_size=val_size,
            random_state=self.random_state,
        )
        return self


    def load_splits_from_json(self, json_path, id_colname):
        splits = io.load_json(json_path)
        test_ids = splits["test"]

        test_rows = self.df[id_colname].isin(test_ids)
        train_rows = ~self.df[id_colname].isin(test_ids)

        self.X_test = self.X.loc[test_rows]
        self.y_test = self.y.loc[test_rows]
        self.X_train = self.X.loc[train_rows]
        self.y_train = self.y.loc[train_rows]

        train_ids = splits["train"]
        self.n_splits = len(train_ids)
        for train_fold_ids, val_fold_ids in train_ids.values():

            train_fold_rows = self.df[id_colname].isin(train_fold_ids)
            val_fold_rows = self.df[id_colname].isin(val_fold_ids)

            self.X_train_fold.append(self.X.loc[train_fold_rows])
            self.X_val_fold.append(self.X.loc[val_fold_rows])
            self.y_train_fold.append(self.y.loc[train_fold_rows])
            self.y_val_fold.append(self.y.loc[val_fold_rows])
        return self

    def save_splits_to_json(json_path):
        """TBD"""
        pass

    def split_dataset_temporal(self, test_size=0.2, val_size=0.2):
        """TBD"""
        pass

    def cross_validation_split_test_from_column(
        self, column_name, test_value, n_splits=5
    ):
        """
        Splits into train and test according to `column_name`, then creates k-fold CV splitter on the train.
        Args:
            column_name: column to be used for train-test split
            test_value: value in the `column_name` indicating case should be test set
            n_splits: number of CV splits
        Returns:
            self.cv_split_generator: CV splitter
        """
        (
            self.X_train,
            self.X_test,
            self.y_train,
            self.y_test,
        ) = self.split_train_test_from_column(column_name, test_value)
        kf = StratifiedKFold(
            n_splits=n_splits, shuffle=True, random_state=self.random_state
        )
        self.cv_split_generator = kf.split(self.X_train, self.y_train)
        self.cv_splits = list(self.cv_split_generator)
        self.create_cross_validation_labels()
        return self

    def standardize_features(self):
        self.X_train[self.X_train.columns] = self.scaler.fit_transform(self.X_train)
        if self.X_val is None:
            print("X_val not set. Leaving out.")
        else:
            self.X_val[self.X_val.columns] = self.scaler.transform(self.X_val)
        self.X_test[self.X_test.columns] = self.scaler.transform(self.X_test)
        self.X[self.X.columns] = self.scaler.transform(self.X)
        return self

    def standardize_features_cross_validation(self):
        self.X_train_fold[self.X_train_fold.columns] = self.scaler.fit_transform(
            self.X_train_fold
        )
        self.X_val_fold[self.X_val_fold.columns] = self.scaler.transform(
            self.X_val_fold
        )

    def inverse_standardize(self, X):
        X[X.columns] = self.scaler.inverse_transform(X)
        return X

    def select_features(self, k=10):
        if self.X_train is None:
            raise ValueError("Split the data into training, validation and test first.")
        else:
            self.feature_selector = SelectKBest(f_classif, k=k)
            self.feature_selector.fit(self.X_train, self.y_train)
            cols = self.feature_selector.get_support(indices=True)
            self.X_train = self.X_train.iloc[:, cols]
            if self.X_val is not None:
                self.X_val = self.X_val.iloc[:, cols]
            self.X_test = self.X_test.iloc[:, cols]
            self.X = self.X.iloc[:, cols]
            self.best_features = self.X.columns
            print(f"Selected features: {self.best_features.values}")

    def select_features_cross_validation(self):
        if self.X_train is None:
            raise ValueError("Split the data into training and test first.")
        else:
            feature_selector = SelectKBest(f_classif, k=10)
            feature_selector.fit(self.X_train_fold, self.y_train_fold)
            cols = feature_selector.get_support(indices=True)
            self.X_train_fold = self.X_train_fold.iloc[:, cols]
            self.X_val_fold = self.X_val_fold.iloc[:, cols]

    def boxplot_by_class(self):
        """
        Plot the distributions of the selected features by the label class.
        """
        features = self.best_features
        nrows, ncols, figsize = get_subplots_dimensions(len(features))
        fig = make_subplots(rows=nrows, cols=ncols)
        xlabels = ["Positive" if label == 1 else "Negative" for label in self.y_test]
        xlabels = np.array(xlabels)
        # X_test = self.inverse_standardize(self.X_test)
        for i, feature in enumerate(features):
            y = self.X_test[feature]
            _, p_val = wilcoxon_unpaired(
                y[xlabels == "Negative"], y[xlabels == "Positive"]
            )
            fig.add_trace(
                go.Box(y=y, x=xlabels, name=f"{feature} p={p_val}"),
                row=i // ncols + 1,
                col=i % ncols + 1,
            )
        fig.update_layout(title_text=f"Selected features:")
        # fig.show()
        # fig.write_html(self.result_dir / "boxplot.html")
        return fig



import setuptools
from setuptools import setup

# Load packages from requirements.txt
with open("requirements-min.txt") as file:
    required_packages = [ln.strip() for ln in file.readlines()]

setup(
    name="classrad",
    packages=setuptools.find_packages(),
    include_package_data=True,
    install_requires=required_packages,
)



import numpy as np
from sklearn import model_selection
from collections import OrderedDict
from classrad.utils import io


def split_train_test(
    ids, labels, test_size=0.2, shuffle=True, random_state=42
):
    """
    Split data into training and test set.
    :param ids: list of ids
    :param labels: list of labels
    :return: train_ids, test_ids, train_labels, test_labels
    """
    # Split data into training and test set
    ids_train, ids_test, y_train, y_test = model_selection.train_test_split(
        ids,
        labels,
        test_size=test_size,
        stratify=labels,
        shuffle=shuffle,
        random_state=random_state,
    )
    return ids_train, ids_test, y_train, y_test


def split_cross_validation(
    ids, labels, n_splits=5, shuffle=True, random_state=42
):
    """
    Split data into n folds for cross-validation
    """
    ids = np.array(ids)
    ids_split = OrderedDict()
    cv = model_selection.StratifiedKFold(
        n_splits=n_splits, shuffle=shuffle, random_state=random_state
    )
    split_indices = list(cv.split(ids, labels))
    for i in range(len(split_indices)):
        train_indices, test_indices = list(split_indices[i])
        train_ids = ids[train_indices].tolist()
        test_ids = ids[test_indices].tolist()
        ids_split[f"fold_{i}"] = (train_ids, test_ids)
    return ids_split


def split_full_dataset(
    ids,
    labels,
    result_dir,
    test_size=0.2,
    n_splits=5,
    shuffle=True,
    random_state=42,
):
    """
    Split data into test and training. Divide training into n folds for
    cross-validation.
    Saves the splits as json.
    """
    ids_train, ids_test, y_train, y_test = split_train_test(
        ids,
        labels,
        test_size=test_size,
        shuffle=shuffle,
        random_state=random_state,
    )
    ids_split = {}
    ids_split["test"] = ids_test
    ids_split["train"] = split_cross_validation(
        ids_train, y_train, n_splits=n_splits, random_state=random_state
    )
    save_path = result_dir / "splits.json"
    io.save_json(ids_split, save_path)
    return ids_split



from pathlib import Path
import streamlit as st

from extractor import StreamlitFeatureExtractor
import utils
from template_utils import radiomics_params
from classrad.config import config


def show():
    with st.sidebar:
        st.write(
            """
            Expected input:
                CSV file with with absolute paths to the image and the mask
                for each case.
        """
        )
    path_df = utils.load_df("Choose a CSV file with paths:")
    st.write(path_df)
    # path_df.replace("", np.nan, inplace=True)
    col1, col2 = st.columns(2)
    colnames = path_df.columns.tolist()
    with col1:
        image_col = st.selectbox("Path to image", colnames)
    with col2:
        mask_col = st.selectbox("Path to segmentation", colnames)
    path_df.dropna(subset=[image_col, mask_col], inplace=True)
    radiomics_params()

    out_path = Path(config.RESULT_DIR) / "features_test.csv"
    num_threads = st.slider(
        "Number of threads", min_value=1, max_value=8, value=1
    )
    start_extraction = st.button("Run feature extraction")

    if start_extraction:
        progressbar = st.progress(0)
        extractor = StreamlitFeatureExtractor(
            df=path_df.head(2),
            out_path=out_path,
            image_col=image_col,
            mask_col=mask_col,
            #  extraction_params=extraction_params,
            verbose=False,
            num_threads=num_threads,
        )
        feature_df = extractor.extract_features(progressbar=progressbar)
        st.success(
            f"Done! Features saved in your result directory ({out_path})"
        )
        feature_colnames = [
            col
            for col in feature_df.columns
            if col.startswith(("original", "wavelet", "shape"))
        ]
        feature_df[feature_colnames] = feature_df[feature_colnames].astype(
            float
        )
        st.write(feature_df)


if __name__ == "__main__":
    show()



import logging
import os
import shutil
from pathlib import Path
import radiomics
import SimpleITK as sitk
import streamlit as st
import utils
from radiomics import featureextractor
from template_utils import radiomics_params_voxelbased
from classrad.config import config


input_dir = Path(config.INPUT_DIR)
result_dir = Path(config.RESULT_DIR)


def show():
    """Shows the sidebar components for the template
    and returns user inputs as dict."""
    with st.sidebar:
        load_test_data = st.checkbox("Load test data to the input directory")
    if load_test_data:
        utils.load_test_data(input_dir)
    filelist = []
    # Find all files in input directory ending with '.nii.gz' or '.nrrd'
    for fpath in input_dir.rglob("*.nii.gz"):
        filelist.append(str(fpath))
    for fpath in input_dir.rglob("*.nii"):
        filelist.append(str(fpath))
    for fpath in input_dir.rglob("*.nrrd"):
        filelist.append(str(fpath))
    filelist = [fpath for fpath in filelist if "results" not in fpath]
    st.write("""Files found in your input directory:""")
    st.write(filelist)

    col1, col2 = st.columns(2)
    with col1:
        image_path = st.text_input("Paste here the path to the image:")
        image_path = image_path.strip('"')
        if os.path.isfile(image_path):
            st.success("Image found!")
    with col2:
        seg_path = st.text_input("Paste here the path to the segmentation:")
        seg_path = seg_path.strip('"')
        if os.path.isfile(seg_path):
            st.success("Segmentation found!")
    col1, col2 = st.columns(2)
    with col1:
        output_dirname = st.text_input(
            "Give this extraction some ID to easily find the results:"
        )
        if output_dirname:
            maps_output_dir = result_dir / output_dirname
            if utils.dir_nonempty(maps_output_dir):
                st.warning("This ID already exists and has some data!")
            else:
                maps_output_dir.mkdir(parents=True, exist_ok=True)
                st.success(f"Maps will be saved in {maps_output_dir}")
    extraction_params = radiomics_params_voxelbased()
    start_extraction = st.button("Get feature maps!")
    if start_extraction:
        assert output_dirname, "You need to assign an ID first! (see above)"
        assert image_path, "You need to provide an image path!"
        assert seg_path, "You need to provide a segmentation path!"
        utils.save_yaml(
            extraction_params, maps_output_dir / "extraction_params.yaml"
        )
        shutil.copyfile(image_path, maps_output_dir / "image.nii.gz")
        shutil.copyfile(seg_path, maps_output_dir / "segmentation.nii.gz")
        with st.spinner("Extracting and saving feature maps..."):
            radiomics.setVerbosity(logging.INFO)
            extractor = featureextractor.RadiomicsFeatureExtractor(
                extraction_params
            )
            feature_vector = extractor.execute(
                image_path, seg_path, voxelBased=True
            )
            for feature_name, feature_value in feature_vector.items():
                if isinstance(feature_value, sitk.Image):
                    save_path = maps_output_dir / f"{feature_name}.nii.gz"
                    save_path = str(save_path)
                    sitk.WriteImage(feature_value, save_path)
        st.success(
            f"Done! Feature maps and configuration saved in {maps_output_dir}"
        )
        if config.IS_DEMO:
            zip_name = f"{output_dirname}.zip"
            zip_save_path = result_dir / zip_name
            utils.zip_directory(str(maps_output_dir), str(zip_save_path))
            with open(str(zip_save_path), "rb") as fp:
                st.download_button(
                    "Download results",
                    data=fp,
                    file_name=zip_name,
                    mime="application/zip",
                )


if __name__ == "__main__":
    show()



import json
import os
import tempfile
import classrad

CONFIG_DIR = os.path.dirname(__file__)
TEST_DATA_DIR = os.path.join(
    os.path.dirname(os.path.dirname(classrad.__file__)),
    "tests",
    "testing_data",
)
IS_DEMO = False
if "INPUT_DIR" in os.environ:
    INPUT_DIR = os.environ["INPUT_DIR"]
else:
    INPUT_DIR = tempfile.mkdtemp()
    IS_DEMO = True
if "RESULT_DIR" in os.environ:
    RESULT_DIR = os.environ["RESULT_DIR"]
else:
    RESULT_DIR = tempfile.mkdtemp()


PARAM_DIR = os.path.join(CONFIG_DIR, "pyradiomics_params")
PRESETS = {"default CT": "default_feature_map.yaml"}
with open(os.path.join(CONFIG_DIR, "pyradiomics_feature_names.json")) as f:
    PYRADIOMICS_FEATURE_NAMES = json.load(f)

AVAILABLE_CLASSIFIERS = [
    "Random Forest",
    "AdaBoost",
    "Logistic Regression",
    "SVM",
    "Gaussian Process Classifier",
    "XGBoost",
]

SEED = 1234



import json
from pathlib import Path

import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns
from lofo import Dataset, LOFOImportance, plot_importance
from classrad.utils.io import load_json
from classrad.utils.statistics import wilcoxon_unpaired
from classrad.utils.visualization import get_subplots_dimensions
from sklearn.model_selection import GridSearchCV


class Trainer:
    def __init__(
        self,
        dataset,
        models,
        result_dir,
        meta_colnames,
        num_features=10,
        n_jobs=1,
        random_state=None,
    ):
        self.dataset = dataset
        self.models = models
        self.result_dir = result_dir
        self.meta_colnames = meta_colnames
        self.num_features = num_features
        self.n_jobs = n_jobs
        self.random_state = random_state
        self.model_names = [model.classifier_name for model in models]
        self.result_df = None
        self.test_indices = None

        self.result_dir.mkdir(parents=True, exist_ok=True)

    def init_result_df(self):
        self.result_df = self.dataset.df[self.meta_colnames].copy()
        self.result_df["test"] = 0
        self.test_indices = self.dataset.X_test.index.values
        self.result_df.loc[self.test_indices, "test"] = 1
        self.result_df["cv_split"] = -1
        for i in range(self.dataset.n_splits):
            self.result_df.loc[self.dataset.X_val_fold[i].index, "cv_split"] = i

    def train_cross_validation(self):
        """
        Evaluate the models.
        """
        self.init_result_df()
        # Feature standardization and selection
        self.dataset.standardize_features()
        self.dataset.select_features(k=self.num_features)

        for model in self.models:
            model_name = model.classifier_name
            pred_colname = f"{model_name}_pred"
            pred_proba_colname = f"{model_name}_pred_proba"
            self.result_df[pred_colname] = -1
            self.result_df[pred_proba_colname] = -1

            print(f"Training and infering model: {model_name}")

            optimizer = HyperparamOptimizer(
                dataset=self.dataset,
                model=model,
                param_dir=self.result_dir / "optimal_params",
            )
            model = optimizer.load_or_tune_hyperparameters()

            for i in range(self.dataset.n_splits):
                print(f"Evaluating fold: {i}")

                X_train_fold = self.dataset.X_train_fold[i]
                y_train_fold = self.dataset.y_train_fold[i]
                X_val_fold = self.dataset.X_val_fold[i]
                # Fit and predict
                model.fit(X_train_fold, y_train_fold)
                y_pred_fold, y_pred_proba_fold = model.predict_label_and_proba(
                    X_val_fold
                )
                # Write results
                fold_indices = X_val_fold.index
                self.result_df.loc[fold_indices, pred_colname] = y_pred_fold

                self.result_df.loc[fold_indices, pred_proba_colname] = y_pred_proba_fold

            model.fit(self.dataset.X_train, self.dataset.y_train)
            y_pred_test, y_pred_proba_test = model.predict_label_and_proba(
                self.dataset.X_test
            )
            self.result_df.loc[self.test_indices, pred_colname] = y_pred_test
            self.result_df.loc[
                self.test_indices, pred_proba_colname
            ] = y_pred_proba_test
        self.save_results()

        return self

    def save_results(self):
        df_name = f"predictions_{self.dataset.task_name}.csv"
        self.result_df.to_csv(self.result_dir / df_name, index=False)
        return self

    def plot_feature_importance(self, model, ax=None):
        """
        Plot importance of features for a single model
        Args:
            model [MLClassifier] - classifier
            ax (optional) - pyplot axes object
        """
        model_name = model.classifier_name
        try:
            importances = model.feature_importance()
            importance_df = pd.DataFrame(
                {"feature": self.dataset.X_train.columns, "importance": importances}
            )
            sns.barplot(x="feature", y="importance", data=importance_df, ax=ax)
            ax.tick_params(axis="both", labelsize="x-small")
            ax.set_ylabel("Feature importance")
            ax.set_title(model_name)
            ax.set_xticklabels(ax.get_xticklabels(), rotation=30, ha="right")
        except Exception:
            print(f"For {model_name} feature importance cannot be calculated.")

        return self

    def plot_feature_importance_all(self, title=None):
        """
        Plot the feature importance for all models.
        """
        nrows, ncols, figsize = get_subplots_dimensions(len(self.models))
        fig, axs = plt.subplots(nrows=nrows, ncols=ncols, figsize=figsize)
        for i, model in enumerate(self.models):
            ax = fig.axes[i]
            self.plot_feature_importance(model, ax=ax)
        if title:
            fig.suptitle(title)
        else:
            fig.suptitle(f"Feature Importance for {self.dataset.task_name}")
        fig.tight_layout()
        fig.savefig(
            self.result_dir / "feature_importance.png", bbox_inches="tight", dpi=100
        )
        plt.show()

        return self

    def plot_lofo_importance(self, model):
        dataset = Dataset(
            df=self.dataset.df,
            target=self.target,
            features=self.dataset.best_features,
        )
        lofo_imp = LOFOImportance(
            dataset, model=model.classifier, scoring="neg_mean_squared_error"
        )
        importance_df = lofo_imp.get_importance()
        plot_importance(importance_df, figsize=(12, 12))
        plt.tight_layout()
        plt.show()


class HyperparamOptimizer:
    def __init__(self, dataset, model, param_dir):
        self.dataset = dataset
        self.model = model
        self.param_dir = Path(param_dir)
        self.param_grid = None

    def get_grid_RandomForest(self):
        n_estimators = [200, 600, 1000]
        max_features = ["auto", "sqrt"]
        max_depth = [10, 50, None]
        min_samples_split = [2, 5, 10]
        min_samples_leaf = [1, 4]
        bootstrap = [True, False]

        self.param_grid = {
            "n_estimators": n_estimators,
            "max_features": max_features,
            "max_depth": max_depth,
            "min_samples_split": min_samples_split,
            "min_samples_leaf": min_samples_leaf,
            "bootstrap": bootstrap,
        }
        return self

    def get_grid_XGBoost(self):
        self.param_grid = {
            "learning_rate": [0.05, 0.10, 0.20, 0.30],
            "max_depth": [2, 4, 8, 12, 15],
            "min_child_weight": [1, 3, 7],
            "gamma": [0.0, 0.1, 0.3],
            "colsample_bytree": [0.3, 0.5, 0.7],
        }
        return self

    def get_grid_LogReg(self):
        self.param_grid = {
            "C": [0.001, 0.01, 0.1, 1, 10, 100, 1000],
            "penalty": ["l1", "l2", "elasticnet", "none"],
        }
        return self

    def get_grid_SVM(self):
        cs = [0.001, 0.01, 0.1, 1, 10]
        gammas = [0.001, 0.01, 0.1, 1]
        kernels = ["rbf"]
        self.param_grid = {"kernel": kernels, "C": cs, "gamma": gammas}
        return self

    def save_params(self, params):
        self.param_dir.mkdir(exist_ok=True)
        save_path = Path(self.param_dir) / (self.model.classifier_name + ".json")
        io.save_json(save_path, params)

    def load_params(self):
        param_path = self.param_dir / (self.model.classifier_name + ".json")
        print(f"Loading parameters from: {param_path}")
        if param_path.exists():
            optimal_params = io.load_json(param_path)
            self.model.set_params(optimal_params)
        else:
            raise FileNotFoundError(
                "No param file found. Run \
                                    `tune_hyperparameters` first."
            )

        return self

    def update_model_params_grid_search(self):
        if self.param_grid is None:
            raise ValueError("First select param grid!")
        else:
            param_searcher = GridSearchCV(
                estimator=self.model.classifier,
                param_grid=self.param_grid,
                scoring="roc_auc",
                cv=self.dataset.cv_splits,
                verbose=0,
                n_jobs=-1,
            )
            rs = param_searcher.fit(X=self.dataset.X_train, y=self.dataset.y_train)
            optimal_params = rs.best_params_
            print(f"Best params for {self.model.classifier_name}: {optimal_params}")
            self.model.set_params(optimal_params)
            self.save_params(optimal_params)

            return self

    def get_param_grid(self):
        model_name = self.model.classifier_name
        if model_name == "Random Forest":
            self.get_grid_RandomForest()
        elif model_name == "XGBoost":
            self.get_grid_XGBoost()
        elif model_name == "Logistic Regression":
            self.get_grid_LogReg()
        elif model_name == "SVM":
            self.get_grid_SVM()
        else:
            return ValueError(
                f"Hyperparameter tuning for {model_name} not implemented!"
            )

        return self

    def tune_hyperparameters(self):
        try:
            self.get_param_grid()
            self.update_model_params_grid_search()
        except Exception:
            pass

        return self

    def load_or_tune_hyperparameters(self):
        try:
            self.load_params()
        except Exception:
            print("Params couldn't be loaded. Starting hyperparameter tuning...")
            self.tune_hyperparameters()

        return self.model
